"""
yadr
~~~~

The core of the :mod:`yadr` package.
"""
from argparse import ArgumentParser
from importlib.resources import open_text
from pathlib import Path
from typing import Optional

from yadr import maps as m
from yadr.encode import Encoder
from yadr.lex import Lexer
from yadr.model import CompoundResult, DiceMapping, Result, TokenInfo
from yadr.parser import dice_map, Parser


# Execute YADN.
def roll(
    yadn: str,
    yadn_out: bool = False,
    dice_map: Optional[dict[str, DiceMapping]] = None
) -> None | Result | CompoundResult:
    """Execute a string of :ref:`YADN` to roll dice.

    :param yadn: A string of :ref:`YADN` that defines the die roll to
        execute.
    :param yadn_out: (Optional.) Whether the output should be in native
        Python objects or :ref:`YADN` notation. The default is native
        Python objects.
    :param dice_map: (Optional.) A dictionary of maps for transforming
        the value rolled. See :ref:`dice_maps` for details.
    :return: The result depends on the details of the die roll.
    :rtype: None, Result, or CompoundResult

    Usage::

        >>> import yadr
        >>>
        >>> yadr.roll('3d6')                        # doctest: +SKIP
        16

    The specific result will depend on the :ref:`YADN` being executed.
    In the example above, it will be an integer in the range of three to
    eighteen that is created by generating three random integers in the
    range of one to six.
    """
    # Get the default dice maps and add any passed into the roll.
    if not dice_map:
        dice_map = {}
    default_maps = get_default_maps()
    default_maps.update(dice_map)

    # Lex the YADN into tokens for parsing.
    lexer = Lexer()
    tokens = lexer.lex(yadn)

    # Parse and execute the YADN tokens.
    parser = Parser()
    parser.dice_map = default_maps
    result: None | Result | CompoundResult = parser.parse(tokens)

    # If needed, encode the result into YADN and return the result.
    if yadn_out:
        encoder = Encoder()
        result = encoder.encode(result)
    return result


# Utility.
def read_file(loc: str | Path) -> str:
    """Read test from a file.

    :param loc: The file system location of the file.
    :return: A :class:str object.
    :rtype: str
    """
    path = Path(loc)
    with open(path) as fh:
        contents = fh.read()
    return contents


def parse_map(yadn: str) -> dict[str, DiceMapping]:
    """Parse the contents of a dice mapping file."""
    if ';' in yadn:
        yadn_parts = yadn.split(';')
        dice_map = {}
        for part in yadn_parts:
            assert ';' not in part
            dice_map.update(parse_map(part))
        return dice_map

    mlexer = m.Lexer()
    mparser = m.Parser()
    tokens = mlexer.lex(yadn)
    name, value = mparser.parse(tokens)
    return {name: value, }


# Command parsing.
def add_dice_map(loc: str) -> dict[str, DiceMapping]:
    """Load the dice-maps from a given file.

    :param loc: The location of the file of dice mappings to load.
    :return: None.
    :rtype: NoneType

    Usage::

        >>> from yadr import add_dice_map
        >>>
        >>> path = 'tests/data/__test_dice_map.txt'
        >>> add_dice_map(path)              # doctest: +NORMALIZE_WHITESPACE
        {'spam': {1: 'eggs', 2: 'bacon', 3: 'eggs', 4: 'tomato'}, 'fudge':
        {1: '-', 2: '', 3: '+'}}
    """
    yadn = read_file(loc)
    return parse_map(yadn)


def get_default_maps() -> dict[str, DiceMapping]:
    """Get the default dice maps.

    :return: The default dice maps as a :class:`dict` of :class:`dict`
        objects.
    :rtype: dict

    Usage::

        >>> from yadr.yadr import get_default_maps
        >>>
        >>> get_default_maps()              # doctest: +ELLIPSIS
        {'sweote boost': {1: '',...
    """
    default_file = open_text('yadr.data', 'dice_maps.yadn')
    default_maps_yadn = default_file.read()
    default_file.close()
    return parse_map(default_maps_yadn)


def list_dice_maps() -> str:
    """Get the list of the default dice maps.

    :return: A :class:`str` object.
    :rtype: str

    Usage::

        >>> from yadr import list_dice_maps
        >>>
        >>> list_dice_maps()                # doctest: +ELLIPSIS
        'sweote boost...
    """
    dice_map = get_default_maps()
    maps_ = '\n'.join(dice_map)
    return maps_


def parse_cli() -> None:
    """Parse command line options."""
    # Stand up the parser.
    p = ArgumentParser(
        description='Execute YADN syntax to roll dice.',
        prog='yadr'
    )

    # Define the command line arguments.
    p.add_argument(
        'yadn',
        help='A string of YADN describing the die roll.',
        action='store',
        nargs='?',
        type=str
    )
    p.add_argument(
        '--list_dice_maps', '-l',
        help='List the names of the default dice maps.',
        action='store_true'
    )
    p.add_argument(
        '--add_dice_map', '-m',
        help='Load the dice mappings at the given file location.',
        nargs=1,
        action='store',
        type=str
    )

    # Parse and execute the command.
    args = p.parse_args()
    result = ''
    dice_map = {}
    if args.add_dice_map:
        dice_map = add_dice_map(args.add_dice_map[0])
    elif args.list_dice_maps:
        result = list_dice_maps()
    if args.yadn:
        raw_result = roll(args.yadn, True, dice_map)
        result = str(raw_result)
    print(result)



"""
test_lex
~~~~~~~~

Unit tests for the dice notation lexer.
"""
import unittest as ut

from tests.common import BaseTests
from yadr import lex
from yadr import model as m


# Symbol test cases.
class ASOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.AS_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_addition(self):
        """Given a basic addition equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 15),
            (lex.Token.AS_OPERATOR, '+'),
            (lex.Token.NUMBER, 3),
        )
        data = '15+3'
        self.lex_test(exp, data)

    def test_basic_addition_with_spaces(self):
        """Given a basic addition equation containing whitespace,
        return the tokens that represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 15),
            (lex.Token.AS_OPERATOR, '+'),
            (lex.Token.NUMBER, 3),
        )
        data = ' 15 + 3 '
        self.lex_test(exp, data)

    def test_basic_subtraction(self):
        """Given a basic subtraction equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 200),
            (lex.Token.AS_OPERATOR, '-'),
            (lex.Token.NUMBER, 10),
        )
        data = '200-10'
        self.lex_test(exp, data)


class BooleanTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.BOOLEAN
    allowed = [
        m.Token.CHOICE_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_boolean_true(self):
        """Lex a boolean."""
        exp = (
            (lex.Token.BOOLEAN, True),
        )
        data = 'T'
        self.lex_test(exp, data)

    def test_boolean_false(self):
        """Lex a boolean."""
        exp = (
            (lex.Token.BOOLEAN, True),
        )
        data = 'T'
        self.lex_test(exp, data)


class ChoiceTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.CHOICE_OPERATOR
    allowed = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_choice(self):
        """Lex a choice operator."""
        exp = (
            (lex.Token.BOOLEAN, True),
            (lex.Token.CHOICE_OPERATOR, '?'),
            (lex.Token.QUALIFIER, 'spam'),
            (lex.Token.OPTIONS_OPERATOR, ':'),
            (lex.Token.QUALIFIER, 'eggs'),
        )
        data = 'T?"spam":"eggs"'
        self.lex_test(exp, data)


class ComparisonOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.COMPARISON_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_equal(self):
        """Lex equal."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '=='),
            (lex.Token.NUMBER, 20),
        )
        data = '21==20'
        self.lex_test(exp, data)

    def test_basic_greater_than(self):
        """Lex greater than."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '>'),
            (lex.Token.NUMBER, 20),
        )
        data = '21>20'
        self.lex_test(exp, data)

    def test_basic_greater_than_or_equal(self):
        """Lex greater than or equal."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '>='),
            (lex.Token.NUMBER, 20),
        )
        data = '21>=20'
        self.lex_test(exp, data)

    def test_basic_greater_than_whitspace(self):
        """Lex greater than."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '>'),
            (lex.Token.NUMBER, 20),
        )
        data = '21 > 20'
        self.lex_test(exp, data)

    def test_basic_less_than(self):
        """Lex greater than."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '<'),
            (lex.Token.NUMBER, 20),
        )
        data = '21<20'
        self.lex_test(exp, data)

    def test_basic_less_than_or_equal(self):
        """Lex less than or equal."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '<='),
            (lex.Token.NUMBER, 20),
        )
        data = '21<=20'
        self.lex_test(exp, data)

    def test_basic_not_equal(self):
        """Lex not equal."""
        exp = (
            (lex.Token.NUMBER, 21),
            (lex.Token.COMPARISON_OPERATOR, '!='),
            (lex.Token.NUMBER, 20),
        )
        data = '21!=20'
        self.lex_test(exp, data)


class DiceOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.DICE_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_concat(self):
        """Given a basic concat equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'dc'),
            (lex.Token.NUMBER, 10),
        )
        data = '20dc10'
        self.lex_test(exp, data)

    def test_basic_die(self):
        """Given a basic die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'd'),
            (lex.Token.NUMBER, 10),
        )
        data = '20d10'
        self.lex_test(exp, data)

    def test_basic_exploding_die(self):
        """Given a basic exploding die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'd!'),
            (lex.Token.NUMBER, 10),
        )
        data = '20d!10'
        self.lex_test(exp, data)

    def test_basic_keep_high_die(self):
        """Given a basic die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'dh'),
            (lex.Token.NUMBER, 10),
        )
        data = '20dh10'
        self.lex_test(exp, data)

    def test_basic_keep_low_die(self):
        """Given a basic die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'dl'),
            (lex.Token.NUMBER, 10),
        )
        data = '20dl10'
        self.lex_test(exp, data)

    def test_basic_wild_die(self):
        """Given a basic die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.DICE_OPERATOR, 'dw'),
            (lex.Token.NUMBER, 10),
        )
        data = '20dw10'
        self.lex_test(exp, data)


class ExOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.EX_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_exponentiation(self):
        """Given a basic exponentiation equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.EX_OPERATOR, '^'),
            (lex.Token.NUMBER, 10),
        )
        data = '20^10'
        self.lex_test(exp, data)


class GroupingTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.GROUP_CLOSE
    allowed = [
        m.Token.AS_OPERATOR,
        m.Token.MD_OPERATOR,
        m.Token.EX_OPERATOR,
        m.Token.DICE_OPERATOR,
        m.Token.GROUP_CLOSE,
        m.Token.POOL_OPERATOR,
        m.Token.ROLL_DELIMITER,
        m.Token.POOL_GEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_parentheses(self):
        """Given a statement containing parenthesis, return the
        tokenized equation.
        """
        exp = (
            (lex.Token.GROUP_OPEN, '('),
            (lex.Token.NUMBER, 32),
            (lex.Token.AS_OPERATOR, '-'),
            (lex.Token.NUMBER, 5),
            (lex.Token.GROUP_CLOSE, ')'),
            (lex.Token.MD_OPERATOR, '*'),
            (lex.Token.NUMBER, 21),
        )
        data = '(32-5)*21'
        self.lex_test(exp, data)

    def test_parentheses_with_whitespace(self):
        """Given a statement containing parenthesis and whitespace,
        return the tokenized equation.
        """
        exp = (
            (lex.Token.GROUP_OPEN, '('),
            (lex.Token.NUMBER, 32),
            (lex.Token.AS_OPERATOR, '-'),
            (lex.Token.NUMBER, 5),
            (lex.Token.GROUP_CLOSE, ')'),
            (lex.Token.MD_OPERATOR, '*'),
            (lex.Token.NUMBER, 21),
        )
        data = '( 32 - 5 ) * 21'
        self.lex_test(exp, data)


class GroupOpenTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.GROUP_OPEN
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]


class MapTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.MAP
    allowed = [
        m.Token.ROLL_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_map(self):
        """Given a statement containing a map, return the tokenized
        dice mapping.
        """
        exp = ((
            m.Token.MAP,
            (
                'name',
                {
                    1: "none",
                    2: "success",
                    3: "success",
                    4: "success success",
                }
            )
        ),)
        yadn = '{"name"=1:"none",2:"success",3:"success",4:"success success"}'
        self.lex_test(exp, yadn)


class MappingOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.MAPPING_OPERATOR
    allowed = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_mapping_operator(self):
        """Lex a mapping operator."""
        exp = (
            (lex.Token.NUMBER, 3),
            (lex.Token.MAPPING_OPERATOR, 'm'),
            (lex.Token.QUALIFIER, 'spam'),
        )
        data = '3m"spam"'
        self.lex_test(exp, data)


class MDOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.MD_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_division(self):
        """Given a basic division equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.MD_OPERATOR, '/'),
            (lex.Token.NUMBER, 10),
        )
        data = '20/10'
        self.lex_test(exp, data)

    def test_basic_exponentiation(self):
        """Given a basic exponentiation equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.EX_OPERATOR, '^'),
            (lex.Token.NUMBER, 10),
        )
        data = '20^10'
        self.lex_test(exp, data)

    def test_basic_modulo(self):
        """Given a basic modulo equation, return the tokens
        that represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 2),
            (lex.Token.MD_OPERATOR, '%'),
            (lex.Token.NUMBER, 10),
        )
        data = '2%10'
        self.lex_test(exp, data)

    def test_basic_multiplication(self):
        """Given a basic multiplication equation, return the tokens
        that represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 2),
            (lex.Token.MD_OPERATOR, '*'),
            (lex.Token.NUMBER, 10),
        )
        data = '2*10'
        self.lex_test(exp, data)


class NumberTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.NUMBER
    allowed = [
        m.Token.AS_OPERATOR,
        m.Token.COMPARISON_OPERATOR,
        m.Token.EX_OPERATOR,
        m.Token.DICE_OPERATOR,
        m.Token.GROUP_CLOSE,
        m.Token.MAPPING_OPERATOR,
        m.Token.MD_OPERATOR,
        m.Token.POOL_GEN_OPERATOR,
        m.Token.ROLL_DELIMITER,
        m.Token.WHITESPACE,
    ]

    # Allowed next symbol.
    def test_number_cannot_follow_number(self):
        """Numbers cannot follow numbers."""
        # Expected values.
        exp_ex = ValueError
        exp_msg = '4 cannot follow a NUMBER.'

        # Test data and state.
        data = '3 4'
        lexer = lex.Lexer()

        # Run test and determine the result.
        with self.assertRaisesRegex(exp_ex, exp_msg):
            _ = lexer.lex(data)


class OptionsOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.OPTIONS_OPERATOR
    allowed = [
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_basic_options_operator(self):
        """Lex choice options."""
        exp = (
            (lex.Token.QUALIFIER, 'spam'),
            (lex.Token.OPTIONS_OPERATOR, ':'),
            (lex.Token.QUALIFIER, 'eggs'),
        )
        data = '"spam":"eggs"'
        self.lex_test(exp, data)


class PoolDegenerationOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.POOL_DEGEN_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_count_successes(self):
        """Given a basic count successes statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_DEGEN_OPERATOR, 'ns'),
            (lex.Token.NUMBER, 5),
        )
        data = '[5,1,9]ns5'
        self.lex_test(exp, data)

    def test_basic_count_successes_with_botch(self):
        """Given a basic count successes with botches statement, return
        the tokens in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_DEGEN_OPERATOR, 'nb'),
            (lex.Token.NUMBER, 5),
        )
        data = '[5,1,9]nb5'
        self.lex_test(exp, data)

    def test_count_successes_before_group(self):
        """Groups can follow pool degeneration operators."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_DEGEN_OPERATOR, 'ns'),
            (lex.Token.GROUP_OPEN, '('),
            (lex.Token.NUMBER, 3),
            (lex.Token.AS_OPERATOR, '+'),
            (lex.Token.NUMBER, 2),
            (lex.Token.GROUP_CLOSE, ')'),
        )
        data = '[5,1,9]ns(3+2)'
        self.lex_test(exp, data)

    def test_count_successes_before_unary_pool_degen(self):
        """Unary pool degens can follow pool degeneration operators."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_DEGEN_OPERATOR, 'ns'),
            (lex.Token.U_POOL_DEGEN_OPERATOR, 'N'),
            (lex.Token.POOL, (5, 1, 9)),
        )
        data = '[5,1,9]nsN[5,1,9]'
        self.lex_test(exp, data)

    def test_count_successes_before_operator(self):
        """Operators cannot occur after pool degen operators."""
        # Expected values.
        exp_ex = ValueError
        exp_msg = '\\+ cannot follow a POOL_DEGEN_OPERATOR.'

        # Test data and state.
        data = '[5,1,9]ns+'
        lexer = lex.Lexer()

        # Run test and determine results.
        with self.assertRaisesRegex(exp_ex, exp_msg):
            _ = lexer.lex(data)


class PoolGenerationOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.POOL_GEN_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_dice_pool(self):
        """Given a basic die equation, return the tokens that
        represent the equation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.POOL_GEN_OPERATOR, 'g'),
            (lex.Token.NUMBER, 10),
        )
        data = '20g10'
        self.lex_test(exp, data)

    def test_basic_expolding_pool(self):
        """Given a basic pool generation, return the tokens that
        represent the generation.
        """
        exp = (
            (lex.Token.NUMBER, 20),
            (lex.Token.POOL_GEN_OPERATOR, 'g!'),
            (lex.Token.NUMBER, 10),
        )
        data = '20g!10'
        self.lex_test(exp, data)


class PoolOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.POOL_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_pool_keep_above(self):
        """Given a basic pool keep above statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pa'),
            (lex.Token.NUMBER, 2),
        )
        data = '[5,1,9]pa2'
        self.lex_test(exp, data)

    def test_basic_pool_keep_below(self):
        """Given a basic pool keep below statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pb'),
            (lex.Token.NUMBER, 2),
        )
        data = '[5,1,9]pb2'
        self.lex_test(exp, data)

    def test_basic_pool_cap(self):
        """Cap the maximum value in a pool."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pc'),
            (lex.Token.NUMBER, 7),
        )
        data = '[5,1,9]pc7'
        self.lex_test(exp, data)

    def test_basic_pool_floor(self):
        """Floor the minimum value in a pool."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pf'),
            (lex.Token.NUMBER, 2),
        )
        data = '[5,1,9]pf2'
        self.lex_test(exp, data)

    def test_basic_pool_keep_high(self):
        """Cap the maximum value in a pool."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'ph'),
            (lex.Token.NUMBER, 2),
        )
        data = '[5,1,9]ph2'
        self.lex_test(exp, data)

    def test_basic_pool_keep_low(self):
        """Cap the maximum value in a pool."""
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pl'),
            (lex.Token.NUMBER, 2),
        )
        data = '[5,1,9]pl2'
        self.lex_test(exp, data)

    def test_basic_pool_modulo(self):
        """Given a basic pool modulo statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'p%'),
            (lex.Token.NUMBER, 5),
        )
        data = '[5,1,9]p%5'
        self.lex_test(exp, data)

    def test_basic_pool_remove(self):
        """Given a basic pool remove statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.POOL, (5, 1, 9)),
            (lex.Token.POOL_OPERATOR, 'pr'),
            (lex.Token.NUMBER, 5),
        )
        data = '[5,1,9]pr5'
        self.lex_test(exp, data)


class PoolTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.POOL
    allowed = [
        m.Token.GROUP_CLOSE,
        m.Token.POOL_OPERATOR,
        m.Token.POOL_DEGEN_OPERATOR,
        m.Token.ROLL_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_pool(self):
        """A pool of dice."""
        exp = ((
            lex.Token.POOL,
            (5, 1, 9),
        ),)
        data = '[5,1,9]'
        self.lex_test(exp, data)

    def test_pool_with_whitespace(self):
        """A pool of dice that has whitespace."""
        exp = ((
            lex.Token.POOL,
            (5, 1, 9),
        ),)
        data = '[ 5 , 1 , 9 ]'
        self.lex_test(exp, data)


class QualifierTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.QUALIFIER
    allowed = [
        m.Token.OPTIONS_OPERATOR,
        m.Token.ROLL_DELIMITER,
        m.Token.WHITESPACE,
    ]

    def test_quotation_marks(self):
        """Given a statement containing quotation marks, return the
        tokenized equation.
        """
        exp = (
            (lex.Token.QUALIFIER, 'spam'),
        )
        data = '"spam"'
        self.lex_test(exp, data)


class ResultsRollTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.ROLL_DELIMITER
    allowed = [
        m.Token.BOOLEAN,
        m.Token.GROUP_OPEN,
        m.Token.MAP_OPEN,
        m.Token.MAP,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.QUALIFIER,
        m.Token.QUALIFIER_DELIMITER,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_roll_delimiter(self):
        """Given a statement containing parenthesis, return the
        tokenized equation.
        """
        exp = (
            (lex.Token.NUMBER, 2),
            (lex.Token.DICE_OPERATOR, 'd'),
            (lex.Token.NUMBER, 10),
            (lex.Token.ROLL_DELIMITER, ';'),
            (lex.Token.NUMBER, 5),
            (lex.Token.DICE_OPERATOR, 'd'),
            (lex.Token.NUMBER, 10),
        )
        data = '2d10;5d10'
        self.lex_test(exp, data)

    def test_roll_delimiter_whitespace(self):
        """Given a statement containing parenthesis, return the
        tokenized equation.
        """
        exp = (
            (lex.Token.NUMBER, 2),
            (lex.Token.DICE_OPERATOR, 'd'),
            (lex.Token.NUMBER, 10),
            (lex.Token.ROLL_DELIMITER, ';'),
            (lex.Token.NUMBER, 5),
            (lex.Token.DICE_OPERATOR, 'd'),
            (lex.Token.NUMBER, 10),
        )
        data = '2d10 ; 5d10'
        self.lex_test(exp, data)


class UnaryPoolDegenerationOperatorTestCase(BaseTests.LexTokenTestCase):
    token = m.Token.U_POOL_DEGEN_OPERATOR
    allowed = [
        m.Token.GROUP_OPEN,
        m.Token.NEGATIVE_SIGN,
        m.Token.NUMBER,
        m.Token.POOL,
        m.Token.POOL_OPEN,
        m.Token.U_POOL_DEGEN_OPERATOR,
        m.Token.WHITESPACE,
    ]

    def test_basic_pool_concatente(self):
        """Given a basic pool concatenate statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.U_POOL_DEGEN_OPERATOR, 'C'),
            (lex.Token.POOL, (3, 1, 7))
        )
        data = 'C[3,1,7]'
        self.lex_test(exp, data)

    def test_basic_pool_count(self):
        """Given a basic pool count statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.U_POOL_DEGEN_OPERATOR, 'N'),
            (lex.Token.POOL, (3, 1, 7))
        )
        data = 'N[3,1,7]'
        self.lex_test(exp, data)

    def test_basic_pool_count_with_space(self):
        """Given a basic pool count statement with white space, return
        the tokens in the statement.
        """
        exp = (
            (lex.Token.U_POOL_DEGEN_OPERATOR, 'N'),
            (lex.Token.POOL, (3, 1, 7))
        )
        data = 'N [3,1,7]'
        self.lex_test(exp, data)

    def test_basic_pool_sum(self):
        """Given a basic pool count statement, return the tokens
        in the statement.
        """
        exp = (
            (lex.Token.U_POOL_DEGEN_OPERATOR, 'S'),
            (lex.Token.POOL, (3, 1, 7))
        )
        data = 'S[3,1,7]'
        self.lex_test(exp, data)


# Roll test case.
class StartRollTestCase(BaseTests.LexTestCase):
    def test_operator_cannot_start_roll(self):
        """An operator cannot start an expression."""
        # Expected values.
        exp_ex = ValueError
        exp_msg = 'Cannot start with \\+.'

        # Test data and state.
        data = '+2'
        lexer = lex.Lexer()

        # Run test and determine the result.
        with self.assertRaisesRegex(exp_ex, exp_msg):
            _ = lexer.lex(data)

    def test_operator_cannot_start_roll_whitespace(self):
        """An operator cannot start an expression."""
        # Expected values.
        exp_ex = ValueError
        exp_msg = 'Cannot start with \\+.'

        # Test data and state.
        data = ' +2'
        lexer = lex.Lexer()

        # Run test and determine the result.
        with self.assertRaisesRegex(exp_ex, exp_msg):
            _ = lexer.lex(data)


# Order of operations test case.
class OrderOfOperationsTestCase(BaseTests.LexTestCase):
    def test_negative_number(self):
        """Tokenize a number that starts with a negative sign."""
        exp = ((lex.Token.NUMBER, -24),)
        data = '-24'
        self.lex_test(exp, data)

    def test_negative_number_after_operator(self):
        """Tokenize a number that starts with a negative sign."""
        exp = (
            (lex.Token.NUMBER, 3),
            (lex.Token.AS_OPERATOR, '+'),
            (lex.Token.NUMBER, -24),
        )
        data = '3+-24'
        self.lex_test(exp, data)

    def test_negative_number_after_operator_with_whitespace(self):
        """Tokenize a number that starts with a negative sign."""
        exp = (
            (lex.Token.NUMBER, 3),
            (lex.Token.AS_OPERATOR, '+'),
            (lex.Token.NUMBER, -24),
        )
        data = '3 + -24'
        self.lex_test(exp, data)



"""
test_parse
~~~~~~~~~~

Unit tests for the yadr.parse module.
"""
import unittest as ut
from unittest.mock import patch

from yadr import parser as p
from yadr.model import Token
from yadr import operator as yo


# Test cases.
class ParserTestCase(ut.TestCase):
    def setUp(self):
        self.parser = p.Parser()

    def tearDown(self):
        self.parser = None

    def parser_test(self, exp, tokens):
        """Return the expected result for the given tokens."""
        act = self.parser.parse(tokens)
        self.assertEqual(exp, act)

    # Tests.
    # Identities
    def test_boolean(self):
        """Parse a boolean."""
        exp = True
        tokens = (
            (Token.BOOLEAN, True),
        )
        self.parser_test(exp, tokens)

    def test_number(self):
        "Parse a number."
        exp = 5
        tokens = (
            (Token.NUMBER, 5),
        )
        self.parser_test(exp, tokens)

    def test_pool(self):
        "Parse a pool."
        exp = (3, 3, 4)
        tokens = (
            (Token.POOL, (3, 3, 4)),
        )
        self.parser_test(exp, tokens)

    def test_quantifier(self):
        """Parse a qualifier."""
        exp = 'spam'
        tokens = (
            (Token.QUALIFIER, 'spam'),
        )
        self.parser_test(exp, tokens)

    # Basic arithmetic.
    def test_addition(self):
        """Parse basic addition."""
        exp = 5
        tokens = (
            (Token.NUMBER, 3),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_division(self):
        "Perform basic division."
        exp = 2
        tokens = (
            (Token.NUMBER, 4),
            (Token.MD_OPERATOR, '/'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_division_rounds_down(self):
        "Division rounds down to the nearest integer (floor division)."
        exp = 2
        tokens = (
            (Token.NUMBER, 5),
            (Token.MD_OPERATOR, '/'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_exponentiation(self):
        "Perform basic division."
        exp = 9
        tokens = (
            (Token.NUMBER, 3),
            (Token.EX_OPERATOR, '^'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_modulo(self):
        "Perform basic modulo."
        exp = 1
        tokens = (
            (Token.NUMBER, 3),
            (Token.MD_OPERATOR, '%'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_multiplication(self):
        "Perform basic multiplication."
        exp = 6
        tokens = (
            (Token.NUMBER, 3),
            (Token.MD_OPERATOR, '*'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_subtraction(self):
        "Perform basic subtraction."
        exp = 1
        tokens = (
            (Token.NUMBER, 3),
            (Token.AS_OPERATOR, '-'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    # Choice and related operators.
    def test_choice(self):
        """Perform choice operation."""
        exp = 'eggs'
        tokens = (
            (Token.BOOLEAN, False),
            (Token.CHOICE_OPERATOR, '?'),
            (Token.QUALIFIER, 'spam'),
            (Token.OPTIONS_OPERATOR, ':'),
            (Token.QUALIFIER, 'eggs'),
        )
        self.parser_test(exp, tokens)

    def test_greater_than(self):
        """Perform greater than comparison."""
        exp = True
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '>'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_greater_than_or_equal(self):
        """Perform greater than or equal comparison."""
        exp = True
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '>='),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_less_than(self):
        """Perform less than comparison."""
        exp = False
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '<'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_less_than_or_equal(self):
        """Perform less than comparison."""
        exp = False
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '<='),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_equal(self):
        """Perform less than comparison."""
        exp = False
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '=='),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_not_equal(self):
        """Perform less than comparison."""
        exp = True
        tokens = (
            (Token.NUMBER, 3),
            (Token.COMPARISON_OPERATOR, '!='),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_options_operator(self):
        """Create a choice options."""
        exp = ('spam', 'eggs')
        tokens = (
            (Token.QUALIFIER, 'spam'),
            (Token.OPTIONS_OPERATOR, ':'),
            (Token.QUALIFIER, 'eggs'),
        )
        self.parser_test(exp, tokens)

    # Dice operators.
    @patch('random.randint')
    def test_concat(self, mock_randint):
        """Concatenate dice rolls."""
        mock_randint.side_effect = [1, 11, 3]
        exp = 113
        tokens = (
            (Token.NUMBER, 3),
            (Token.DICE_OPERATOR, 'dc'),
            (Token.NUMBER, 12),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_die(self, mock_randint):
        """Roll a die."""
        mock_randint.side_effect = [1, 2, 2]
        exp = 5
        tokens = (
            (Token.NUMBER, 3),
            (Token.DICE_OPERATOR, 'd'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_exploding_die(self, mock_randint):
        """Roll an exploding die."""
        mock_randint.side_effect = [1, 2, 4, 2, 2]
        exp = 11
        tokens = (
            (Token.NUMBER, 4),
            (Token.DICE_OPERATOR, 'd!'),
            (Token.NUMBER, 4),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_keep_high_die(self, mock_randint):
        """Roll dice and keep the highest."""
        mock_randint.side_effect = [1, 5, 3]
        exp = 5
        tokens = (
            (Token.NUMBER, 3),
            (Token.DICE_OPERATOR, 'dh'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_keep_low_die(self, mock_randint):
        """Roll dice and keep the lowest."""
        mock_randint.side_effect = [1, 5, 3]
        exp = 1
        tokens = (
            (Token.NUMBER, 3),
            (Token.DICE_OPERATOR, 'dl'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_wild_die(self, mock_randint):
        """Roll dice and keep the lowest."""
        mock_randint.side_effect = [4, 2, 1, 6]
        exp = 13
        tokens = (
            (Token.NUMBER, 4),
            (Token.DICE_OPERATOR, 'dw'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    # Pool generation operators.
    @patch('random.randint')
    def test_dice_pool(self, mock_randint):
        """Roll dice and keep the highest."""
        mock_randint.side_effect = [1, 5, 3]
        exp = (1, 5, 3)
        tokens = (
            (Token.NUMBER, 3),
            (Token.POOL_GEN_OPERATOR, 'g'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_exploding_pool(self, mock_randint):
        """Roll an exploding dice pool."""
        mock_randint.side_effect = [1, 5, 3]
        exp = (1, 5, 3)
        tokens = (
            (Token.NUMBER, 3),
            (Token.POOL_GEN_OPERATOR, 'g!'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    # Unary pool degeneration operators.
    @patch('random.randint')
    def test_pool_concatenate(self, mock_randint):
        """Concatenate the dice in a pool."""
        mock_randint.side_effect = [1, 5, 3]
        exp = 153
        tokens = (
            (Token.U_POOL_DEGEN_OPERATOR, 'C'),
            (Token.NUMBER, (1, 5, 3)),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_pool_count(self, mock_randint):
        """Count the dice in a pool."""
        mock_randint.side_effect = [1, 5, 3]
        exp = 3
        tokens = (
            (Token.U_POOL_DEGEN_OPERATOR, 'N'),
            (Token.NUMBER, (1, 5, 3)),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_pool_sum(self, mock_randint):
        """Count the dice in a pool."""
        mock_randint.side_effect = [1, 5, 3]
        exp = 9
        tokens = (
            (Token.U_POOL_DEGEN_OPERATOR, 'S'),
            (Token.NUMBER, (1, 5, 3)),
        )
        self.parser_test(exp, tokens)

    # Pool degeneration operators.
    def test_count_successes(self):
        """Keep a values that aren't a given number in the pool."""
        exp = 2
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_DEGEN_OPERATOR, 'ns'),
            (Token.NUMBER, 8),
        )
        self.parser_test(exp, tokens)

    def test_count_successes_with_botch(self):
        """Keep a values that aren't a given number in the pool."""
        exp = 1
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_DEGEN_OPERATOR, 'nb'),
            (Token.NUMBER, 8),
        )
        self.parser_test(exp, tokens)

    # Pool operators.
    def test_pool_keep_above(self):
        """Keep a values above a given number in the pool."""
        exp = (8, 9)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pa'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    def test_pool_keep_below(self):
        """Keep a values below a given number in the pool."""
        exp = (5, 1)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pb'),
            (Token.NUMBER, 6),
        )
        self.parser_test(exp, tokens)

    def test_pool_cap(self):
        """Cap the values in a pool at a given value."""
        exp = (5, 7, 1, 7)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pc'),
            (Token.NUMBER, 7),
        )
        self.parser_test(exp, tokens)

    def test_pool_floor(self):
        """Floor the values in a pool at a given value."""
        exp = (8, 8, 8, 9)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pf'),
            (Token.NUMBER, 8),
        )
        self.parser_test(exp, tokens)

    def test_pool_keep_high(self):
        """Keep a number of highest values from the pool."""
        exp = (8, 9)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'ph'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_pool_keep_low(self):
        """Keep a number of highest values from the pool."""
        exp = (5, 1)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pl'),
            (Token.NUMBER, 2),
        )
        self.parser_test(exp, tokens)

    def test_pool_modulo(self):
        """Perform a modulo on each member."""
        exp = (0, 3, 1, 4)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'p%'),
            (Token.NUMBER, 5),
        )
        self.parser_test(exp, tokens)

    def test_pool_remove(self):
        """Keep a values that aren't a given number in the pool."""
        exp = (5, 1, 9)
        tokens = (
            (Token.POOL, (5, 8, 1, 9)),
            (Token.POOL_OPERATOR, 'pr'),
            (Token.NUMBER, 8),
        )
        self.parser_test(exp, tokens)

    # Dice mapping.
    def test_map(self):
        """Parse and store a dice map."""
        # Expected value.
        exp = 'success'

        # Test data and state.
        tokens = (
            (Token.MAP, ('spam', {1: 'none', 2: 'success'})),
        )
        _ = self.parser.parse(tokens)

        # Run test.
        act = self.parser.dice_map['spam'][2]

        # Determine test result.
        self.assertEqual(exp, act)

    # Mapping operators.
    @patch('random.randint')
    def test_map_result(self, mock_randint):
        """Given a map and a roll with a dice map operator, return the
        correct result.
        """
        # Expected value.
        exp = 'success'

        # Test data and state.
        parser = p.Parser()
        mock_randint.side_effect = (3, )
        parser.dice_map['_test_map_result'] = {
            1: 'none',
            2: 'success',
            3: 'success',
            4: 'success success',
        }
        tokens = (
            (Token.NUMBER, 1),
            (Token.DICE_OPERATOR, 'd'),
            (Token.NUMBER, 4),
            (Token.MAPPING_OPERATOR, 'm'),
            (Token.QUALIFIER, '_test_map_result'),
        )

        # Run test.
        act = parser.parse(tokens)

        # Determine test result.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_map_result_pool(self, mock_randint):
        """Given a map and a roll with a dice map operator, return the
        correct result.
        """
        # Expected value.
        exp = ('success', 'none')

        # Test data and state.
        mock_randint.side_effect = (3, 1, )
        self.parser.dice_map['_test_map_result'] = {
            1: 'none',
            2: 'success',
            3: 'success',
            4: 'success success',
        }
        tokens = (
            (Token.NUMBER, 2),
            (Token.DICE_OPERATOR, 'g'),
            (Token.NUMBER, 4),
            (Token.MAPPING_OPERATOR, 'm'),
            (Token.QUALIFIER, '_test_map_result'),
        )

        # Run test.
        act = self.parser.parse(tokens)

        # Determine test result.
        self.assertEqual(exp, act)

    @patch('random.randint')
    # Test rolls and results.
    def test_roll_delimiter(self, mock_randint):
        """Return the result of two rolls."""
        mock_randint.side_effect = (1, 1, 3, 8)
        exp = ((1, 1, 3), 10)
        tokens = (
            (Token.NUMBER, 3),
            (Token.POOL_GEN_OPERATOR, 'g'),
            (Token.NUMBER, 8),
            (Token.ROLL_DELIMITER, ';'),
            (Token.NUMBER, 1),
            (Token.DICE_OPERATOR, 'd'),
            (Token.NUMBER, 10),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 2)
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_roll_delimiter_remove_none_result(self, mock_randint):
        """Return the result of two rolls."""
        mock_randint.side_effect = (2,)
        exp = 'lose'
        tokens = (
            (Token.MAP, (
                'spam',
                {1: "win", 2: "lose"}
            )),
            (Token.ROLL_DELIMITER, ';'),
            (Token.NUMBER, 1),
            (Token.DICE_OPERATOR, 'd'),
            (Token.NUMBER, 2),
            (Token.MAPPING_OPERATOR, 'm'),
            (Token.QUALIFIER, 'spam')
        )
        self.parser_test(exp, tokens)

    # Test order of precedence.
    def test_can_perform_multiple_operations(self):
        """The parser can parse statements with multiple operators."""
        exp = 9
        tokens = (
            (Token.NUMBER, 3),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 2),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 4),
        )
        self.parser_test(exp, tokens)

    def test_exponentiation_before_multiplication(self):
        """Multiplication should occur before addition."""
        exp = 48
        tokens = (
            (Token.NUMBER, 3),
            (Token.MD_OPERATOR, '*'),
            (Token.NUMBER, 2),
            (Token.EX_OPERATOR, '^'),
            (Token.NUMBER, 4),
        )
        self.parser_test(exp, tokens)

    def test_multiplication_before_addition(self):
        """Multiplication should occur before addition."""
        exp = 11
        tokens = (
            (Token.NUMBER, 3),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 2),
            (Token.MD_OPERATOR, '*'),
            (Token.NUMBER, 4),
        )
        self.parser_test(exp, tokens)

    def test_parens_before_multiplication(self):
        """Parentheses should occur before multiplication."""
        exp = 18
        tokens = (
            (Token.NUMBER, 3),
            (Token.MD_OPERATOR, '*'),
            (Token.GROUP_OPEN, '('),
            (Token.NUMBER, 2),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 4),
            (Token.GROUP_CLOSE, ')'),
        )
        self.parser_test(exp, tokens)

    @patch('random.randint')
    def test_pool_generation_happens_before_pool_degeneration(self, mock_ri):
        """Pool generation should happen before pool degeneration
        in order of operations.
        """
        # Expected value.
        exp = 20

        # Test data and state.
        rand_results = [5, 4, 6, 2, 3]
        mock_ri.side_effect = rand_results
        tokens = (
            (Token.U_POOL_DEGEN_OPERATOR, 'S'),
            (Token.NUMBER, 5),
            (Token.POOL_GEN_OPERATOR, 'g'),
            (Token.NUMBER, 6),
        )

        # Run test.
        act = self.parser.parse(tokens)

        # Determine test result.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_unary_pool_degeneration_happens_before_operators(self, mock_ri):
        """Unary pool degeneration should happen before operators
        in order of operations.
        """
        # Expected value.
        exp = 25

        # Test data and state.
        rand_results = [5, 4, 6, 2, 3]
        mock_ri.side_effect = rand_results
        tokens = (
            (Token.U_POOL_DEGEN_OPERATOR, 'S'),
            (Token.NUMBER, 5),
            (Token.POOL_GEN_OPERATOR, 'g'),
            (Token.NUMBER, 6),
            (Token.AS_OPERATOR, '+'),
            (Token.NUMBER, 5),
        )

        # Run test.
        act = self.parser.parse(tokens)

        # Determine test result.
        self.assertEqual(exp, act)

    def test_pool_operation_happens_before_pool_degeneration(self):
        """Pool operations should happen before pool degeneration
        in order of operations.
        """
        # Expected value.
        exp = 1

        # Test data and state.
        tokens = (
            (Token.POOL, (5, 4, 6, 2, 3)),
            (Token.POOL_OPERATOR, 'pb'),
            (Token.NUMBER, 5),
            (Token.POOL_DEGEN_OPERATOR, 'ns'),
            (Token.NUMBER, 5),
        )

        # Run test.
        act = self.parser.parse(tokens)

        # Determine test result.
        self.assertEqual(exp, act)



"""
test_encoder
~~~~~~~~~~~~

Unit tests for the yadr.encoder module.
"""
import unittest as ut

from yadr import encode as e
from yadr import model as m


# Encoder test case.
class EncoderTestCase(ut.TestCase):
    def setUp(self):
        self.encoder = e.Encoder()

    def tearDown(self):
        self.encoder = None

    def test_boolean(self):
        """An bool becomes a string containing the YADN boolean."""
        exp = 'T'
        data = True
        act = self.encoder.encode(data)
        self.assertEqual(exp, act)

    def test_compound_result(self):
        """A CompoundResult becomes a string of roll delimited
        values.
        """
        exp = '[1, 1, 3, 8]; 3'
        data = m.CompoundResult((
            (1, 1, 3, 8),
            3
        ))
        act = self.encoder.encode(data)
        self.assertEqual(exp, act)

    def test_number(self):
        """An int becomes a string containing the YADN number."""
        exp = '3'
        data = 3
        act = self.encoder.encode(data)
        self.assertEqual(exp, act)

    def test_pool(self):
        """A tuple of integers becomes a string containing the
        YADN pool.
        """
        exp = '[1, 1, 3, 8]'
        data = (1, 1, 3, 8)
        act = self.encoder.encode(data)
        self.assertEqual(exp, act)

    def test_qualifier(self):
        """A string becomes a double-quoted string.
        """
        exp = '"spam"'
        data = 'spam'
        act = self.encoder.encode(data)
        self.assertEqual(exp, act)



"""
test_maps
~~~~~~~~~

Unittests for the yadr.maps module.
"""
import unittest as ut

from tests.common import BaseTests
from yadr import maps
from yadr.model import Token


# Lexing test cases.
class KVDelimiterTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.KV_DELIMITER
    allowed = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
        Token.QUALIFIER,
        Token.QUALIFIER_DELIMITER,
        Token.WHITESPACE,
    ]

    def test_kv_delimiter(self):
        """Given a key-value delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
        )
        yadn = '{"spam"=1:'
        self.lex_test(exp, yadn)

    def test_kv_delimiter_whitespace(self):
        """Given a key-value delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
        )
        yadn = '{"spam"=1 :'
        self.lex_test(exp, yadn)


class MapCloseTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.MAP_CLOSE
    allowed = []

    def test_map_close(self):
        """Given a map close character, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.MAP_CLOSE, '}'),
        )
        yadn = '{}'
        self.lex_test(exp, yadn)

    def test_map_close_whitespace(self):
        """Given a map close character, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.MAP_CLOSE, '}'),
        )
        yadn = '{ }'
        self.lex_test(exp, yadn)


class MapOpenTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.MAP_OPEN
    allowed = [
        Token.MAP_CLOSE,
        Token.QUALIFIER,
        Token.QUALIFIER_DELIMITER,
        Token.WHITESPACE,
    ]

    def test_map_open(self):
        """Given a map open character, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
        )
        yadn = '{'
        self.lex_test(exp, yadn)


class NameDelimiterTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.NAME_DELIMITER
    allowed = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
        Token.WHITESPACE,
    ]

    def test_name_delimiter(self):
        """Given a name delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '=')
        )
        yadn = '{"spam"='
        self.lex_test(exp, yadn)

    def test_name_delimiter_whitespace(self):
        """Given a name delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '=')
        )
        yadn = '{"spam" ='
        self.lex_test(exp, yadn)


class NegativeSignTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.NEGATIVE_SIGN
    allowed = [
        Token.NUMBER,
    ]

    def test_negative_sign(self):
        """Given a negative sign, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, -1),
        )
        yadn = '{"spam"=-1'
        self.lex_test(exp, yadn)


class NumberTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.NUMBER
    allowed = [
        Token.KV_DELIMITER,
        Token.MAP_CLOSE,
        Token.PAIR_DELIMITER,
    ]

    def test_number(self):
        """Given a number, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
        )
        yadn = '{"spam"=1'
        self.lex_test(exp, yadn)

    def test_number_whitespace(self):
        """Given a number, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
        )
        yadn = '{"spam"= 1'
        self.lex_test(exp, yadn)


class PairDelimiterTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.PAIR_DELIMITER
    allowed = [
        Token.NEGATIVE_SIGN,
        Token.NUMBER,
        Token.WHITESPACE,
    ]

    def test_pair_delimiter(self):
        """Given a pair delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'value'),
            (Token.PAIR_DELIMITER, ','),
        )
        yadn = '{"spam"=1:"value",'
        self.lex_test(exp, yadn)

    def test_pair_delimiter_whitespace(self):
        """Given a pair delimiter, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'value'),
            (Token.PAIR_DELIMITER, ','),
        )
        yadn = '{"spam"=1:"value" ,'
        self.lex_test(exp, yadn)


class QualifierTestCase(BaseTests.MapLexTokenTestCase):
    token = Token.QUALIFIER
    allowed = [
        Token.MAP_CLOSE,
        Token.NAME_DELIMITER,
        Token.PAIR_DELIMITER,
        Token.WHITESPACE,
    ]

    def test_qualifier(self):
        """Given a qualifier, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam')
        )
        yadn = '{"spam"'
        self.lex_test(exp, yadn)

    def test_qualifier_whitespace(self):
        """Given a qualifier, return the proper tokens."""
        exp = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'spam')
        )
        yadn = '{ "spam"'
        self.lex_test(exp, yadn)


# Parsing test cases.
class ParseTestCase(ut.TestCase):
    def setUp(self):
        self.parser = maps.Parser()

    def tearDown(self):
        self.parser = None

    def parser_test(self, exp, tokens):
        act = self.parser.parse(tokens)
        self.assertEqual(exp, act)

    # Test cases.
    def test_parser(self):
        """A basic dice mapping can be parsed."""
        exp = (
            'name',
            {
                1: "none",
                2: "success",
                3: "success",
                4: "success success",
            }
        )
        tokens = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'name'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'none'),
            (Token.PAIR_DELIMITER, ','),
            (Token.NUMBER, 2),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'success'),
            (Token.PAIR_DELIMITER, ','),
            (Token.NUMBER, 3),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'success'),
            (Token.PAIR_DELIMITER, ','),
            (Token.NUMBER, 4),
            (Token.KV_DELIMITER, ':'),
            (Token.QUALIFIER, 'success success'),
            (Token.PAIR_DELIMITER, ','),
            (Token.MAP_CLOSE, '}'),
        )
        self.parser_test(exp, tokens)

    def test_parser_with_numbers(self):
        """A basic dice mapping can be parsed."""
        exp = (
            'name',
            {
                1: -1,
                2: 0,
                3: 1,
            }
        )
        tokens = (
            (Token.MAP_OPEN, '{'),
            (Token.QUALIFIER, 'name'),
            (Token.NAME_DELIMITER, '='),
            (Token.NUMBER, 1),
            (Token.KV_DELIMITER, ':'),
            (Token.NUMBER, -1),
            (Token.PAIR_DELIMITER, ','),
            (Token.NUMBER, 2),
            (Token.KV_DELIMITER, ':'),
            (Token.NUMBER, 0),
            (Token.PAIR_DELIMITER, ','),
            (Token.NUMBER, 3),
            (Token.KV_DELIMITER, ':'),
            (Token.NUMBER, 1),
            (Token.MAP_CLOSE, '}'),
        )
        self.parser_test(exp, tokens)



"""
test_yadr
~~~~~~~~~

Unit tests for the yadr.yadr module.
"""
from copy import deepcopy
from io import StringIO
import sys
import unittest as ut
from unittest.mock import patch

from yadr import yadr


# Test cases.
class ParseCliTestCase(ut.TestCase):
    def setUp(self):
        self.argv_buffer = sys.argv

    def tearDown(self):
        sys.argv = self.argv_buffer

    @patch('sys.stdout', new_callable=StringIO)
    @patch('random.randint')
    def test_yadn(self, mock_randint, mock_stdout):
        """Execute YADN from the command line."""
        # Expected value.
        exp = '11\n'

        # Test data and state.
        sys.argv = ['python -m yadr', '3d6']
        mock_randint.side_effect = (4, 4, 3)

        # Run test.
        yadr.parse_cli()

        # Extract actual result and determine success.
        act = mock_stdout.getvalue()
        self.assertEqual(exp, act)

    @patch('sys.stdout', new_callable=StringIO)
    @patch('random.randint')
    def test_yadn_dice_maps(self, mock_randint, mock_stdout):
        """The -m option followed by a file path should load in the
        dice maps at the given location.
        """
        # Expected value.
        exp = '["+", ""]\n'

        # Test data and state.
        sys.argv = [
            'python -m yadr',
            '2g3m"fudge"',
            '-m',
            'tests/data/__test_dice_map.txt'
        ]
        mock_randint.side_effect = (3, 2)

        # Run test.
        yadr.parse_cli()

        # Extract actual result and determine success.
        act = mock_stdout.getvalue()
        self.assertEqual(exp, act)

    @patch('sys.stdout', new_callable=StringIO)
    @patch('random.randint')
    def test_list_default_dice_maps(self, mock_randint, mock_stdout):
        """The -l option will list the default dice maps."""
        # Expected data and state.
        default_map_loc = 'yadr/data/dice_maps.yadn'
        with open(default_map_loc) as fh:
            lines = fh.readlines()
        lines = [line for line in lines if '=' in line]
        lines = [line.split('"')[1] for line in lines]

        # Expected value.
        exp = '\n'.join(lines) + '\n'

        # Test data and state.
        sys.argv = [
            'python -m yadr',
            '-l'
        ]
        mock_randint.side_effect = (3, 2)

        # Run test.
        yadr.parse_cli()

        # Extract actual result and determine success.
        act = mock_stdout.getvalue()
        self.assertEqual(exp, act)


class RollTestCase(ut.TestCase):
    @patch('random.randint')
    def test_roll_default_dice_maps(self, mock_randint):
        """Execute a YADN string using a default dice map."""
        # Expected value.
        exp = '-'

        # Test data and state.
        mock_randint.side_effect = (1,)
        yadn = '1d3m"fate"'

        # Run test.
        act = yadr.roll(yadn)

        # Determine test results.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_roll(self, mock_randint):
        """Execute a YADN string."""
        # Expected value.
        exp = 11

        # Test data and state.
        mock_randint.side_effect = (4, 4, 3)
        yadn = '3d6'

        # Run test.
        act = yadr.roll(yadn)

        # Determine test results.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_roll_yadn_out(self, mock_randint):
        """Execute a YADN string."""
        # Expected value.
        exp = '11'

        # Test data and state.
        mock_randint.side_effect = (4, 4, 3)
        yadn = '3d6'
        yadn_out = True

        # Run test.
        act = yadr.roll(yadn, yadn_out)

        # Determine test results.
        self.assertEqual(exp, act)



"""
test_operator
~~~~~~~~~~~~~

Dice operators for the `yadr` package.
"""
import unittest as ut
from unittest.mock import patch

from yadr import operator as op


# Choice test cases.
class ChoiceTestCase(ut.TestCase):
    def test_choice_options(self):
        """Generate choice options."""
        # Expected value.
        exp = ('spam', 'eggs')

        # Test data and state.
        a = exp[0]
        b = exp[1]

        # Run test.
        act = op.choice_options(a, b)

        # Determine test result.
        self.assertEqual(exp, act)

    def test_choice(self):
        """Make a choice."""
        # Expected value.
        exp = 'spam'

        # Test data and state.
        boolean = True
        options = ('spam', 'eggs')

        # Run test.
        act = op.choice(boolean, options)

        # Determine test result.
        self.assertEqual(exp, act)


# Dice operation test cases.
class ConcatTestCase(ut.TestCase):
    @patch('random.randint')
    def test_concat(self, mock_randint):
        """Concatenate the least significant digit of the dice."""
        # Expected value.
        exp = 304

        # Test data and state.
        mock_randint.side_effect = (3, 10, 4)
        num = 3
        size = 10

        # Run test.
        act = op.concat(num, size)

        # Determine test result.
        self.assertEqual(exp, act)


class DieTestCase(ut.TestCase):
    def die_test(self, exp, args=None, kwargs=None, seed='spam'):
        """Common test for the die function."""
        if not args:
            args = []
        if not kwargs:
            kwargs = {}
        op._seed(seed)
        act = op.die(*args, **kwargs)
        self.assertEqual(exp, act)

    def test_die(self):
        """Given a number of dice and the size of the die,
        roll that many dice and return the result.
        """
        exp = 4
        kwargs = {
            'num': 1,
            'size': 6,
        }
        seed = 'spam12'
        self.die_test(exp, kwargs=kwargs, seed=seed)


class ExplodingDie(ut.TestCase):
    def exploding_die_test(self, exp, num, size):
        """Common test for the die function."""
        act = op.exploding_die(num, size)
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_exploding_die(self, mock_randint):
        """Given a number of dice and the size of the die,
        roll that many exploding dice and return the result.
        """
        exp = 25
        mock_randint.side_effect = [2, 1, 4, 4, 3, 1, 4, 4, 2]
        num = 5
        size = 4
        self.exploding_die_test(exp, num, size)


class KeepHighDie(ut.TestCase):
    @patch('random.randint')
    def test_keep_high_die(self, mock_randint):
        # Expected value.
        exp = 18

        # Test data and state.
        mock_randint.side_effect = [15, 3, 6, 18, 10]
        num = 5
        size = 20

        # Run test.
        act = op.keep_high_die(num, size)

        # Determine test result.
        self.assertEqual(exp, act)


class KeepLowDie(ut.TestCase):
    @patch('random.randint')
    def test_keep_high_die(self, mock_randint):
        # Expected value.
        exp = 3

        # Test data and state.
        mock_randint.side_effect = [15, 3, 6, 18, 10]
        num = 5
        size = 20

        # Run test.
        act = op.keep_low_die(num, size)

        # Determine test result.
        self.assertEqual(exp, act)


class WildDie(ut.TestCase):
    @patch('random.randint')
    def test_wild_die(self, mock_randint):
        # Expected value.
        exp = 17

        # Test data and state.
        mock_randint.side_effect = [3, 4, 1, 5, 4]
        num = 5
        size = 6

        # Run test.
        act = op.wild_die(num, size)

        # Determine test result.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_wild_die_explodes(self, mock_randint):
        # Expected value.
        exp = 22

        # Test data and state.
        mock_randint.side_effect = [6, 2, 4, 1, 5, 4]
        num = 5
        size = 6

        # Run test.
        act = op.wild_die(num, size)

        # Determine test result.
        self.assertEqual(exp, act)

    @patch('random.randint')
    def test_wild_die_is_one(self, mock_randint):
        # Expected value.
        exp = 0

        # Test data and state.
        mock_randint.side_effect = [1, 4, 1, 5, 4]
        num = 5
        size = 6

        # Run test.
        act = op.wild_die(num, size)

        # Determine test result.
        self.assertEqual(exp, act)


# Pool degeneration test cases.
class PoolConcatenateTestCase(ut.TestCase):
    def test_pool_concatenate(self):
        """Concatenate the members in the pool."""
        exp = 314
        pool = (3, 1, 4)
        act = op.pool_concatenate(pool)
        self.assertEqual(exp, act)


class PoolCountTestCase(ut.TestCase):
    def test_pool_count(self):
        """Count the members in the pool."""
        exp = 3
        pool = (3, 1, 4)
        act = op.pool_count(pool)
        self.assertEqual(exp, act)


class PoolSumTestCase(ut.TestCase):
    def test_pool_count(self):
        """Sum the members in the pool."""
        exp = 8
        pool = (3, 1, 4)
        act = op.pool_sum(pool)
        self.assertEqual(exp, act)


class CountSuccessesTestCase(ut.TestCase):
    def test_count_successes(self):
        """Count the number of values above or equal to a target."""
        # Expected value.
        exp = 5

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        target = 5

        # Run test.
        act = op.count_successes(pool, target)

        # Determine test result.
        self.assertEqual(exp, act)


class CountSuccessesWithBotchesTestCase(ut.TestCase):
    def test_count_successes_with_botch(self):
        """Count the number of values above or equal to a target and
        remove botches.
        """
        # Expected value.
        exp = 3

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        target = 5

        # Run test.
        act = op.count_successes_with_botch(pool, target)

        # Determine test result.
        self.assertEqual(exp, act)


# Pool generation operation test cases.
class DicePoolTestCase(ut.TestCase):
    @patch('random.randint')
    def test_dice_pool(self, mock_randint):
        """Generate a dice pool."""
        # Expected value.
        exp = (3, 5, 1, 10, 8, 4, 3)

        # Test data and state.
        mock_randint.side_effect = exp
        num = 7
        size = 10

        # Run test.
        act = op.dice_pool(num, size)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class ExplodingPoolTestCase(ut.TestCase):
    @patch('random.randint')
    def test_exploding_pool(self, mock_randint):
        """Generate a dice pool."""
        # Expected value.
        exp = (2, 9, 1, 1, 13, 3)

        # Test data and state.
        mock_randint.side_effect = (2, 6, 1, 1, 6, 3, 3, 6, 1)
        num = 6
        size = 6

        # Run test.
        act = op.exploding_pool(num, size)

        # Determine test result.
        self.assertTupleEqual(exp, act)


# Pool operation test cases.
class PoolKeepAbove(ut.TestCase):
    def test_pool_keep_above(self):
        """Keep dice equal to or below above the given value from the
        pool.
        """
        # Expected value.
        exp = (5, 6, 4, 5, 6, 6)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        keep = 4

        # Run test.
        act = op.pool_keep_above(pool, keep)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolKeepBelow(ut.TestCase):
    def test_pool_keep_below(self):
        """Keep dice equal to or below the given value from the pool."""
        # Expected value.
        exp = (1, 2, 4, 1, 3)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        ceiling = 4

        # Run test.
        act = op.pool_keep_below(pool, ceiling)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolCapTestCase(ut.TestCase):
    def test_pool_cap(self):
        """Dice in the pool are capped at the given value."""
        # Expected value.
        exp = (1, 2, 3, 7, 5, 6, 7, 7, 7, 4)

        # Test data and state.
        pool = (1, 2, 3, 10, 5, 6, 7, 8, 9, 4)
        cap = 7

        # Run test.
        act = op.pool_cap(pool, cap)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolFloorTestCase(ut.TestCase):
    def test_pool_floor(self):
        """Dice in the pool are floored at the given value."""
        # Expected value.
        exp = (3, 3, 3, 10, 5, 6, 7, 8, 9, 4)

        # Test data and state.
        pool = (1, 2, 3, 10, 5, 6, 7, 8, 9, 4)
        cap = 3

        # Run test.
        act = op.pool_floor(pool, cap)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolKeepHigh(ut.TestCase):
    def test_pool_keep_high(self):
        """Keep the given number of highest dice from the pool."""
        # Expected value.
        exp = (5, 6, 5, 6, 6)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        keep = 5

        # Run test.
        act = op.pool_keep_high(pool, keep)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolKeepLow(ut.TestCase):
    def test_pool_keep_low(self):
        """Keep the given number of highest dice from the pool."""
        # Expected value.
        exp = (1, 2, 1)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        keep = 3

        # Run test.
        act = op.pool_keep_low(pool, keep)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolModulo(ut.TestCase):
    def test_pool_keep_below(self):
        """Perform a modulo on all members."""
        # Expected value.
        exp = (1, 2, 2, 0, 1, 2, 1, 0, 0, 0)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        divisor = 3

        # Run test.
        act = op.pool_modulo(pool, divisor)

        # Determine test result.
        self.assertTupleEqual(exp, act)


class PoolRemove(ut.TestCase):
    def test_pool_keep_below(self):
        """Keep dice equal to or below the given value from the pool."""
        # Expected value.
        exp = (1, 2, 6, 4, 1, 6, 3, 6)

        # Test data and state.
        pool = (1, 2, 5, 6, 4, 5, 1, 6, 3, 6)
        cut = 5

        # Run test.
        act = op.pool_remove(pool, cut)

        # Determine test result.
        self.assertTupleEqual(exp, act)
